// UrlRewriter - A .NET URL Rewriter module
// Version 1.5
//
// Copyright 2005 Intelligencia
// Copyright 2005 Seth Yates
// 

using System;
using System.IO;
using System.Net;
using System.Xml;
using System.Web;
using System.Web.Caching;
using System.Configuration;
using System.Text.RegularExpressions;
using System.Collections.Specialized;
using Intelligencia.UrlRewriter.Configuration;

namespace Intelligencia.UrlRewriter
{
	/// <summary>
	/// Encapsulates all rewriting information about an individual rewrite request.  This class cannot be inherited.
	/// </summary>
	/// <remarks>
	/// This class cannot be created directly.  It will be provided to actions and conditions
	/// by the framework.
	/// </remarks>
	public sealed class RewriteContext
	{
		/// <summary>
		/// Default constructor.
		/// </summary>
		/// <param name="context">The HTTP Context.</param>
		/// <param name="configuration">The configuration.</param>
		internal RewriteContext(HttpContext context, RewriterConfiguration configuration)
		{
			_configuration = configuration;
			_location = context.Request.RawUrl;
			_method = context.Request.HttpMethod;

			foreach (string key in context.Request.ServerVariables)
			{
				_properties.Add(key, context.Request.ServerVariables[key]);
			}

			foreach (string key in context.Request.Headers)
			{
				_properties.Add(key, context.Request.Headers[key]);
			}

			foreach (string key in context.Request.Cookies)
			{
				_properties.Add(key, context.Request.Cookies[key].Value);
			}
		}

		/// <summary>
		/// The current location being rewritten.
		/// </summary>
		/// <remarks>
		/// This property starts out as Request.RawUrl and is altered by various
		/// rewrite actions.
		/// </remarks>
		public string Location
		{
			get
			{
				return _location;
			}
			set
			{
				_location = value;
			}
		}

		/// <summary>
		/// The request method (GET, PUT, POST, HEAD, DELETE).
		/// </summary>
		public string Method
		{
			get
			{
				return _method;
			}
		}

		/// <summary>
		/// The Processing flag.
		/// </summary>
		private RewriteProcessing Processing
		{
			get
			{
				return _processing;
			}
			set
			{
				_processing = value;
			}
		}

		/// <summary>
		/// The properties for the context, including headers and cookie values.
		/// </summary>
		public NameValueCollection Properties
		{
			get
			{
				return _properties;
			}
		}

		/// <summary>
		/// Output headers.
		/// </summary>
		/// <remarks>
		/// This collection is the collection of headers to add to the response.
		/// For the headers sent in the request, use the <see cref="RewriteContext.Properties">Properties</see> property.
		/// </remarks>
		public NameValueCollection Headers
		{
			get
			{
				return _headers;
			}
		}

		/// <summary>
		/// The status code to send in the response.
		/// </summary>
		public HttpStatusCode StatusCode
		{
			get
			{
				return _statusCode;
			}
			set
			{
				_statusCode = value;
			}
		}

		/// <summary>
		/// Collection of output cookies.
		/// </summary>
		/// <remarks>
		/// This is the collection of cookies to send in the response.  For the cookies
		/// received in the request, use the <see cref="RewriteContext.Properties">Properties</see> property.
		/// </remarks>
		public HttpCookieCollection Cookies
		{
			get
			{
				return _cookies;
			}
		}

		/// <summary>
		/// Last matching pattern from a match (if any).
		/// </summary>
		public Match LastMatch
		{
			get
			{
				return _lastMatch;
			}
			set
			{
				_lastMatch = value;
			}
		}

		/// <summary>
		/// Expands the given input using the last match, properties, maps and transforms.
		/// </summary>
		/// <param name="input">The input to expand.</param>
		/// <returns>The expanded form of the input.</returns>
		public string Expand(string input)
		{
			return RewriterHttpModule.Rewriter.Expand(this, input);
		}

		/// <summary>
		/// Resolves the location to an absolute reference.
		/// </summary>
		/// <param name="location">The application-referenced location.</param>
		/// <returns>The absolute location.</returns>
		public string ResolveLocation(string location)
		{
			if (location.StartsWith("~/"))
			{
				location = location.Replace("~", HttpContext.Current.Request.ApplicationPath);
			}

			return location;
		}

		/// <summary>
		/// The configuration of the rewriter.
		/// </summary>
		public RewriterConfiguration Configuration
		{
			get
			{
				return _configuration;
			}
		}

		private string _method = String.Empty;
		private HttpStatusCode _statusCode = HttpStatusCode.OK;
		private RewriteProcessing _processing = RewriteProcessing.ContinueProcessing;
		private string _location;
		private NameValueCollection _properties = new NameValueCollection();
		private NameValueCollection _headers = new NameValueCollection();
		private HttpCookieCollection _cookies = new HttpCookieCollection();
		private Match _lastMatch = null;
		private RewriterConfiguration _configuration = null;
	}
}
