// UrlRewriter - A .NET URL Rewriter module
// Version 1.5
//
// Copyright 2005 Intelligencia
// Copyright 2005 Seth Yates
// 

using System;
using System.IO;
using System.Xml;
using System.Net;
using System.Web;
using System.Web.Caching;
using System.Configuration;
using System.Collections;
using System.Collections.Specialized;
using System.Text.RegularExpressions;
using Intelligencia.UrlRewriter.Configuration;
using Intelligencia.UrlRewriter.Actions;
using Intelligencia.UrlRewriter.Errors;
using Intelligencia.UrlRewriter.Conditions;
using Intelligencia.UrlRewriter.Utilities;

namespace Intelligencia.UrlRewriter
{
	/// <summary>
	/// The core RewriterEngine class.
	/// </summary>
	internal class RewriterEngine
	{
		private string Reduce(RewriteContext context, StringReader reader)
		{
			string result;
			char ch = (char)reader.Read();
			if (Char.IsDigit(ch))
			{
				string num = ch.ToString();
				if (Char.IsDigit((char)reader.Peek()))
				{
					ch = (char)reader.Read();
					num += ch.ToString();
				}
				if (context.LastMatch != null)
				{
					Group group = context.LastMatch.Groups[Convert.ToInt32(num)];
					if (group != null)
					{
						result = group.Value;
					}
					else
					{
						result = String.Empty;
					}
				}
				else
				{
					result = String.Empty;
				}
			}
			else if (ch == '<')
			{
				StringWriter writer = new StringWriter();

				ch = (char)reader.Read();
				while (ch != '>' && ch != (char)65535)
				{
					if (ch == '$')
					{
						writer.Write(Reduce(context, reader));
					}
					else
					{
						writer.Write(ch);
					}
					ch = (char)reader.Read();
				}

				string expr = writer.GetStringBuilder().ToString();
				if (context.LastMatch != null)
				{
					Group group = context.LastMatch.Groups[expr];
					if (group != null)
					{
						result = group.Value;
					}
					else
					{
						result = String.Empty;
					}
				}
				else
				{
					result = String.Empty;
				}
			}			
			else if (ch == '{')
			{
				StringWriter writer = new StringWriter();
				bool isMap = false;
				bool isFunction = false;

				ch = (char)reader.Read();
				while (ch != '}' && ch != (char)65535)
				{
					if (ch == '$')
					{
						writer.Write(Reduce(context, reader));
					}
					else
					{
						if (ch == ':') isMap = true;
						else if (ch == '(') isFunction = true;
						writer.Write(ch);
					}
					ch = (char)reader.Read();
				}

				string expr = writer.GetStringBuilder().ToString();
				if (isMap)
				{
					Match match = Regex.Match(expr, @"^([^\:]+)\:([^\|]+)(\|(.+))?$");
					string mapName = match.Groups[1].Value;
					string mapArgument = match.Groups[2].Value;
					string mapDefault = match.Groups[4].Value;
					result = context.Configuration.TransformFactory.GetTransform(mapName).ApplyTransform(mapArgument);
					if (result == null)
					{
						result = mapDefault;
					}
				}
				else if (isFunction)
				{
					Match match = Regex.Match(expr, @"^([^\(]+)\(([^\)]+)\)$");
					string functionName = match.Groups[1].Value;
					string functionArgument = match.Groups[2].Value;
					result = context.Configuration.TransformFactory.GetTransform(functionName).ApplyTransform(functionArgument);
				}
				else
				{
					result = context.Properties[expr];
				}
			}
			else
			{
				result = ch.ToString();
			}

			return result;
		}

		public string Expand(RewriteContext context, string input)
		{
			/* replacement :- $n
			 * |	${[a-zA-Z0-9\-]+}
			 * |	${fn( <replacement> )}
			 * |	${<replacement-or-id>:<replacement-or-value>:<replacement-or-value>}
			 * 
			 * replacement-or-id :- <replacement> | <id>
			 * replacement-or-value :- <replacement> | <value>
			 */

			/* $1 - regex replacement
			 * ${propertyname}
			 * ${map-name:value}				map-name is replacement, value is replacement
			 * ${map-name:value|default-value}	map-name is replacement, value is replacement, default-value is replacement
			 * ${fn(value)}						value is replacement
			 */

			StringReader reader = new StringReader(input);
			StringWriter writer = new StringWriter();

			char ch = (char)reader.Read();
			while (ch != (char)65535)
			{
				if ((char)ch == '$')
				{
					writer.Write(Reduce(context, reader));
				}
				else
				{
					writer.Write((char)ch);
				}

				ch = (char)reader.Read();
			}

			return writer.GetStringBuilder().ToString();
		}

		public void Rewrite(string host)
		{
			// Create the context
			RewriterConfiguration config = Configuration;
			RewriteContext context = new RewriteContext(HttpContext.Current, config);

			// Process each rule.
			ProcessRules(context);

			// Verify that the url exists on this server.
			VerifyResultExists(context, host);

			// Rewrite the path if the location has changed.
			if ((context.Location != HttpContext.Current.Request.RawUrl) && ((int)context.StatusCode < 400))
			{
				ChangeLocation(context);
			}
			else if (context.StatusCode != HttpStatusCode.OK)
			{
				HandleError(context);
			}

			// Append any headers defined.
			AppendHeaders(context);

			// Append any cookies defined.
			AppendCookies(context);
		}

		public RewriterConfiguration Configuration
		{
			get
			{
				lock (this)
				{
					RewriterConfiguration configuration = HttpContext.Current.Cache.Get(typeof(RewriterEngine).AssemblyQualifiedName) as RewriterConfiguration;
					if (configuration == null)
					{
						configuration = LoadConfiguration();
					}

					return configuration;
				}
			}
		}

		private void ProcessRules(RewriteContext context)
		{
			IList rewriteRules = context.Configuration.Rules;
			for (int i = 0; i < rewriteRules.Count; i++)
			{
				// If the rule is conditional, ensure the conditions are met.
				IRewriteCondition condition = rewriteRules[i] as IRewriteCondition;
				if (condition == null || condition.IsMatch(context))
				{
					// Execute the action.
					IRewriteAction action = rewriteRules[i] as IRewriteAction;
					action.Execute(context);
				
					// If the action is Stop, then break out of the processing loop
					if (action.Processing == RewriteProcessing.StopProcessing)
					{
						break;
					}
					else if (action.Processing == RewriteProcessing.RestartProcessing)
					{
						// Restart from the first rule.
						i = 0;
					}
				}
			}
		}

		private void VerifyResultExists(RewriteContext context, string host)
		{
			if ((context.Location != HttpContext.Current.Request.RawUrl) && ((int)context.StatusCode < 400))
			{
				Uri uri = new Uri(HttpContext.Current.Request.Url, context.Location);
				if (uri.Host == host)
				{
					string filename = HttpContext.Current.Server.MapPath(uri.AbsolutePath);
					if (!File.Exists(filename))
					{
						context.StatusCode = HttpStatusCode.NotFound;
					}
				}
			}
		}

		private void HandleError(RewriteContext context)
		{
			// Return the status code.
			HttpContext.Current.Response.StatusCode = (int)context.StatusCode;

			// Get the error handler if there is one.
			IRewriteErrorHandler handler = context.Configuration.ErrorHandlers[(int)context.StatusCode] as IRewriteErrorHandler;
			if (handler != null)
			{
				// Execute the error handler.
				handler.HandleError(HttpContext.Current);
			}
		}

		private void ChangeLocation(RewriteContext context)
		{
			HttpContext.Current.Response.StatusCode = (int)context.StatusCode;
			if ((int)context.StatusCode < 300)
			{
				// Successful status if less than 300
				HttpContext.Current.RewritePath(context.Location);
			}
			else
			{
				// Redirection or error if >= 300
				HttpContext.Current.Response.RedirectLocation = context.Location;
			}
		}

		private void AppendHeaders(RewriteContext context)
		{
			foreach (string headerKey in context.Headers)
			{
				HttpContext.Current.Response.AppendHeader(headerKey, context.Headers[headerKey]);
			}
		}

		private void AppendCookies(RewriteContext context)
		{
			foreach (HttpCookie cookie in context.Cookies)
			{
				HttpContext.Current.Response.AppendCookie(cookie);
			}
		}

		private RewriterConfiguration LoadConfiguration()
		{
			XmlNode section = ConfigurationSettings.GetConfig(Constants.RewriterNode) as XmlNode;
			RewriterConfiguration config = null;

			XmlNode filenameNode = section.Attributes.GetNamedItem(Constants.AttrFile);
			if (filenameNode != null)
			{
				string filename = HttpContext.Current.Server.MapPath(filenameNode.Value);
				config = LoadConfigurationFromFile(filename);
				if (config != null)
				{
					CacheDependency fileDependency = new CacheDependency(filename);
					HttpContext.Current.Cache.Add(typeof(RewriterEngine).AssemblyQualifiedName, config, fileDependency, DateTime.Now.AddHours(1.0), TimeSpan.Zero, CacheItemPriority.Default, null);
				}
			}

			if (config == null)
			{
				config = LoadConfigurationFromNode(section);
				HttpContext.Current.Cache.Add(typeof(RewriterEngine).AssemblyQualifiedName, config, null, DateTime.Now.AddHours(1.0), TimeSpan.Zero, CacheItemPriority.Default, null);
			}

			return config;
		}

		private RewriterConfiguration LoadConfigurationFromFile(string filename)
		{
			if (File.Exists(filename))
			{
				XmlDocument document = new XmlDocument();
				document.Load(filename);

				return LoadConfigurationFromNode(document.DocumentElement);
			}

			return null;
		}

		private RewriterConfiguration LoadConfigurationFromNode(XmlNode node)
		{
			RewriterConfigurationReader reader = new RewriterConfigurationReader();
			return (RewriterConfiguration)reader.Read(node);
		}
	}
}
